#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Math::BaseCnv 'cnv';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyFLAT;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;


# John Leech, "A Problem on Strings of Beads", The Mathematical Gazette,
# volume 41, number 338, December 1957, item 2726, pages 277-278.
# https://doi.org/10.2307/3610126
# https://www.cambridge.org/core/journals/mathematical-gazette/article/2726-a-problem-on-strings-of-beads/117D6C88D82AAF4CBEFEC6EEE1F5CBBA
# https://www.jstor.org/stable/3610114
#
# Mentioned in:
#
# Dwight R. Bean, Andrzej Ehrenfeucht, George F. McNulty, "Avoidable
# Patterns in Strings of Symbols".


{
  # count ternary squarefrees
  # A006156 count
  # 3,6,12,18,30,42,60,78,108,144,204,264,342
  #
  # A060688 dissimilars, so up to remapping
  #    = A006156(n+2)/6
  # 1,1,2,3,5,7,10,13,18,24,34,44,57,76,103,133,174,232,305
  # ~/OEIS/A006156.internal.txt
  #
  # dissimilars up to remap and reversal
  # 1,1,2,2,4,4,7,7,11,13,20,23,32,39,56,68,93,118,160,202,274,349,464,591
  # not in OEIS: 13,20,23,32,39,56,68,93,118,160,202,274,349,464,591
  #
  # A012212 palindromes
  #
  # palindromes dissimilars
  # not in OEIS: 1,1,1,2,2,2,2,3,3,4,6,8,10,14,17
  #              1       5        10     13
  #
  my @strs = ('');
  foreach my $len (1 .. 30) {
    @strs = extend_squarefrees(\@strs);
    my @strs = @strs;
    # @strs = uniq_up_to_remapping(\@strs);
    @strs = uniq_up_to_remapping_and_reversal(\@strs);

    # all
    print scalar(@strs),",";

    # # palindromes, only ever odd length
    # if ($len & 1) {
    #   my @palindromes = grep {$_ eq reverse $_} @strs;
    #   print scalar(@palindromes),",";
    # }
  }
  print "\n";
  print "four:\n";
  @strs = sort(list_squarefrees(4));
  foreach my $str (@strs) {
    print "$str\n";
  }
  exit 0;
}

{
  # binary squarefrees, few
  my @strs = ('');
  foreach my $len (1 .. 6) {
    @strs = map { my $old = $_;
                  map { my $new = $_.$old; str_has_square($new) ? () : ($new) } '0','1'
                } @strs;
    print join(', ',sort @strs),"\n";
    # print scalar(@strs),",";
  }
  print "\n";
  exit 0;
}
{
  # search uniform cyclic morphisms

  # len=1  0 palindrome [1,1]
  # len=1  1 palindrome (DUP) [1,1]
  # len=1  2 palindrome (DUP) [1,1]
  # len=13  0121021201210 palindrome [1,1]
  # len=13  0212012102120 palindrome (DUP) [1,1]
  # len=13  1020120210201 palindrome (DUP) [1,1]
  # len=13  1202102012021 palindrome (DUP) [1,1]
  # len=13  2010210120102 palindrome (DUP) [1,1]
  # len=13  2101201021012 palindrome (DUP) [1,1]
  # len=17  01202120102120210 palindrome [1,1]
  # len=17  02101210201210120 palindrome (DUP) [1,1]
  # len=17  10212021012021201 palindrome (DUP) [1,1]
  # len=17  12010201210201021 palindrome (DUP) [1,1]
  # len=17  20121012021012102 palindrome (DUP) [1,1]
  # len=17  21020102120102012 palindrome (DUP) [1,1]
  # len=18  012021020102120210 [1,1]
  # len=18  012021201020120210 [1,1]
  # len=18  021012010201210120 (DUP) [1,1]
  # len=18  021012102010210120 (DUP) [1,1]

  my @strs = ('');
  foreach my $len (1 .. 20) {
    @strs = extend_squarefrees(\@strs);
    my %seen;
    foreach my $str (sort @strs) {
      my $morphism = str_to_cyclic_morphism($str);
      my $is_squarefree           = is_squarefree_morphism($morphism) ? 1 : 0;
      my $is_squarefree_by_expand = is_squarefree_morphism_by_expand($morphism) ? 1 : 0;
      if ($is_squarefree || $is_squarefree_by_expand) {
        my $is_palindrome = $str eq reverse($str);
        print "len=$len  $str",
          ($is_palindrome ? " palindrome" : ""),
          (exists $seen{$str} ? " (DUP)" : ""),
          " [$is_squarefree,$is_squarefree_by_expand]",
          "\n";
        @seen{str_remaps($str)} = (); # hash slice
      }
      if ($is_squarefree != $is_squarefree_by_expand) {
        die;
      }
    }
  }
  exit 0;
}

sub uniq_up_to_remapping {
  my ($aref) = @_;
  my %seen;
  return grep { my $want = ! exists $seen{$_};
                @seen{str_remaps($_)} = ();
                $want } sort(@$aref);
}
CHECK {
  my @array = uniq_up_to_remapping(['0','1','02','01']);
  join(',',@array) eq '0,01' or die;
}

sub uniq_up_to_remapping_and_reversal {
  my ($aref) = @_;
  my %seen;
  return grep { my $want = ! exists $seen{$_};
                @seen{map {$_,scalar(reverse $_)} str_remaps($_)} = ();
                $want } sort(@$aref);
}
CHECK {
  my @array = uniq_up_to_remapping_and_reversal(['0','02','20','01']);
  join(',',@array) eq '0,01' or die;
}

sub str_to_cyclic_morphism {
  my ($A) = @_;
  my $B = $A; $B =~ tr/012/120/;
  my $C = $A; $C =~ tr/012/201/;
  return [$A,$B,$C];
}

# $morphism is an arrayref of 3 strings.
# $str is a string, expand it according to the morphism.
sub expand {
  my ($str, $morphism) = @_;
  $str =~ s{.}{$morphism->[$&]}eg;
  return $str;
}

sub str_has_square {
  my ($str) = @_;
  return ($str =~ /(.+)\1/);
}

# Return a list of strings which are squarefree of length $len.
sub list_squarefrees {
  my ($len) = @_;
  ### list_squarefrees(): $len
  my @ret = ('');
  foreach (1 .. $len) {
    @ret = extend_squarefrees(\@ret);
  }
  return @ret;
}
CHECK {
  foreach my $str (list_squarefrees(4)) { length($str) == 4 or die; }
}

# $aref is an arrayref of strings which must be all the same length and
# should be squarefree.
# Return a list of strings with a new symbol at the start when doing so
# remains squarefree.
my @extend_res;
BEGIN { @extend_res = (qr/z/); }  # half length=0 never square, so a no-match
sub extend_squarefrees {
  my ($aref) = @_;
  ### extend_squarefrees(): $aref
  my $h = (length($aref->[0])+1) >> 1;    # len=1 -> half=1
  my $re = ($extend_res[$h] //= qr/^(.{1,$h})\1/);   # matches a square
  return map { my $old = $_;
               map { my $new = $_.$old; $new =~ $re ? () : ($new) } '0','1','2'
             } @$aref;
}
CHECK {
  my @array = extend_squarefrees(['']);
  join(',',@array) eq '0,1,2' or die;
  @array = extend_squarefrees(['0']);
  join(',',@array) eq '10,20' or die;
}

sub str_remaps {
  my ($A) = @_;
  my $B = $A; $B =~ tr/012/021/;
  my $C = $A; $C =~ tr/012/102/;
  my $D = $A; $D =~ tr/012/120/;
  my $E = $A; $E =~ tr/012/201/;
  my $F = $A; $F =~ tr/012/210/;
  return ($A,$B,$C,$D,$E,$F);
}
CHECK {
  my @strs = str_remaps('012');
  my %seen; @seen{@strs} = ();
  scalar(keys %seen) == 6 or die;
  join(',',@strs) eq join(',',
                          '012',
                          '021',
                          '102',
                          '120',
                          '201',
                          '210') or die;
}

{
  # Max Crochemore, "Sharp Characterizations of Squarefree Morphisms",
  # Theor. Comput. Sci., volume 18, 1982, pages 221-226.
  #
  # Little tightening of Thue is it?, so that:
  #
  #     * For uniform, images of all length=3 squarefrees are squarefree
  #       <=> morphism is squarefree.
  #
  #     * For any, images of all length=5 squarefrees are squarefree
  #       <=> morphism is squarefree.

  my @try;
  BEGIN { @try = list_squarefrees(8); }

  sub is_squarefree_morphism {
    my ($morphism) = @_;
    foreach my $str (@try) {
      my $exp = expand($str,$morphism);
      if ($exp =~ /(.+)\1/) {
        return 0;
      }
    }
    return 1;
  }
}
sub is_squarefree_morphism_by_expand {
  my ($morphism) = @_;
  my $str = '0';
  foreach (1 .. 3) {   # bit slow to go to 5
    $str = expand($str,$morphism);
    if (str_has_square($str)) {
      return 0;
    }
  }
  return 1;
}

{
  # search non-cyclic uniform morphisms
  foreach my $len (11) {
    my $count = 0;
    my @arrays = list_squarefrees($len);
    foreach my $i (0 .. $#arrays) {
      my $A = $arrays[$i];
      foreach my $j (0 .. $#arrays) {
        my $B = $arrays[$j];
        foreach my $k (0 .. $#arrays) {
          my $C = $arrays[$k];
          if (is_uniform_squarefree($A,$B,$C)) {
            $count++;
            my $Astr = join('',@$A);
            my $Bstr = join('',@$B);
            my $Cstr = join('',@$C);
            print "found len=$len $Astr",
              ($Astr eq reverse($Astr) ? " palindrome" : ""),"\n";
            print "      len=$len $Bstr",
              ($Bstr eq $Astr ? " =A" : ""),
              ($Bstr eq reverse($Bstr) ? " palindrome" : ""),"\n";
            print "      len=$len $Cstr",
              ($Cstr eq $Astr ? " =A" : ""),
              ($Cstr eq $Bstr ? " =B" : ""),
              ($Cstr eq reverse($Cstr) ? "palindrome" : ""),"\n";
          }
        }
      }
    }
    print "count $count of len=$len\n";
  }
  exit 0;
}

{
  # search uniform cyclic morphisms, possibly symmetric
  my $symmetric = 0;
  my $try;
  $try = sub {
    my ($aref, $pos) = @_;
    ### try: join('', map {$_//'.'} @$aref)
    ### $pos
    my $rpos = ($symmetric ? $#$aref - $pos : $pos);
    if ($pos > $#$aref || $rpos < $pos) {
      if (is_uniform_cyclic_squarefree($aref)) {
        print join('',@$aref),"\n";
      }
      return;
    }
    my @new = @$aref;
    foreach my $s (0,1,2) {
      $new[$pos] = $new[$rpos] = $s;
      my @part = @new[0..$pos];
      ### part: join('', map {$_//'.'} @part)
      if (aref_has_square(\@part)) {
        ### has square, skip ...
        next;
      }
      $try->(\@new, $pos+1);
    }
  };

  {
    my @Leech = (0,1,2,1,0,2,1,2,0,1,2,1,0);
    my $got = is_uniform_cyclic_squarefree(\@Leech);
    ### $got;
  }

  foreach my $len (2..13) {
    my @array = ((undef) x $len);
    $try->(\@array, 0);
  }
  print "\n";

  exit 0;
}



{

  # GP-DEFINE  { my(table=[0,1,2,1,0,2,1,2,0,1,2,1,0]);
  # GP-DEFINE    Leech(n) = my(v=digits(n,#table));
  # GP-DEFINE               sum(i=1,#v,table[v[i]+1]) % 3;
  # GP-DEFINE  }
  # vector(20,n,n--; Leech(n)+1)
  # not in OEIS: 0,1,2,1,0,2,1,2,0,1,2,1,0,1,2,0,2,1,0,2,0,1,2,0,2,1,2,0,1,0,2,1,0,1
  # not in OEIS: 1,2,3,2,1,3,2,3,1,2,3,2,1,2,3,1,3,2,1,3,1,2,3,1,3,2,3,1,2,1,3,2,1,2
  # GP-Test  13^3 == 2197
  # GP-Test  2*13^3 == 4394
  # GP-Test  3*13^3 == 6591
  # GP-Test  13^4 == 28561
  # my(v=OEIS_samples("A337005")); vector(#v,n,n--; Leech(n)) == v
  # my(g=OEIS_bfile_gf("A337005")); g == sum(n=0,poldegree(g), Leech(n)*x^n)
  # my(g=OEIS_bfile_gf("A337005")); poldegree(g)

  # ~/OEIS/b337005.txt
  # system("rm -f /tmp/b337005.txt"); \
  # for(n=0,3*13^3-1, write("/tmp/b337005.txt",Str(n" "Leech(n)))); \
  # system("ls -l /tmp/b337005.txt");
  #               1    3      6      9   11
  my @table = ([0,1,2,1,0, 2,1,2, 0,1,2,1,0],
               [1,2,0,2,1, 0,2,0, 1,2,0,2,1],
               [2,0,1,0,2, 1,0,1, 2,0,1,0,2]);
  my $width = scalar(@{$table[0]});
  print "width $width\n";
  my @h = (0);
  foreach (1 .. 2) {
    @h = map {@{$table[$_]}} @h;
  }
  # @h = map {$_+1} @h;
  print join(',',@h),"\n";
  my $str = join('',@h);
  if ($str =~ /(.+)\1/) {
    print "square $&\n";
  } else {
    print "square free\n";
  }

  my $by_digits = sub {
    my ($n) = @_;
    my $ret = 0;
    foreach my $digit (digit_split_lowtohigh($n,13)) {
      $ret += $table[0]->[$digit];
    }
    return $ret % 3;
  };
  foreach my $n (0 .. $#h) {
    $by_digits->($n) == $h[$n] or die "n=$n";
  }

  my $f = FLAT::DFA->new;
  $f->add_states(3);
  $f->MyFLAT::set_name('high to low');
  $f->set_starting(0);
  $f->set_accepting(0);
  foreach my $from (0 .. 2) {
    foreach my $digit (0 .. $width-1) {
      my $to = $table[$from]->[$digit];
      $f->add_transition($from,$to, $digit);
    }
  }
  # $f->MyFLAT::view;
  # $f = $f->MyFLAT::reverse
  #   ->MyFLAT::set_name('low to high')
  #   ->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;

  exit 0;
}

# GP-DEFINE  { my(table=[0,1,2,1,0,-1,-2,-1,0,1,2,1,0]);
# GP-DEFINE    Leech_signed(n) = my(v=digits(n,#table));
# GP-DEFINE               sum(i=1,#v,table[v[i]+1]);
# GP-DEFINE  }
# vector(13^3,n,n--; Leech_signed(n)) % 3 == \
# vector(13^3,n,n--; Leech(n))
#
# vector(20,n,n--; Leech_signed(n))
# not in OEIS: 0, 1, 2, 1, 0, -1, -2, -1, 0, 1, 2, 1, 0, 1, 2, 3, 2, 1, 0, -1
