#!/usr/bin/perl -w

# Copyright 2019 Kevin Hyde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use FindBin;
use List::Util 'min','max';

use lib $FindBin::Bin;
use MyDragon 'bitlength','Epred','EpredL','EpredR','A','dir';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

my $level_lo = 4;
my $level_aggregate = 7;
my $level_hi = 11;

my $A_type = 'AL';
foreach my $arg (@ARGV) {
  if ($arg =~ /^A/) {
    $A_type = $arg;
  } else {
  }
}
my %A_type_to_Epred_func = (A => \&Epred,
                            AL => \&EpredL,
                            AR => \&EpredR);
my $Epred_func = $A_type_to_Epred_func{$A_type};

print "  % generated by tools/$FindBin::Script $A_type\n";

my $mathit_A_type = length($A_type)==1 ? $A_type : "\\mathit{$A_type}";
print "  \\newcommand\\MyAtype{$mathit_A_type}\n";

my @An = (0);
foreach my $n (0 .. 2**$level_hi) {
  $An[$n+1] = $An[$n] + $Epred_func->($n);
}

sub func {
  my ($n) = @_;
  return sprintf '%.6f', $An[$n] / $n;

  # if ($type eq 'ratio') {
  # } elsif ($type eq 'diff') {
  #   return $An[$n] - 1/2*$n;
  # } else {
  #   die;
  # }
}

sub n_to_x {
  my ($n) = @_;
  if ($n==0) { return 0; }
  my $k = bitlength($n);
  my $high = 2**($k-1);
  my $r = $n - $high;
  return ($r ? "$k+$r/$high" : $k);
}
n_to_x(0) == 0 or die;
n_to_x(1) == 1 or die;
n_to_x(2) == 2 or die;
n_to_x(4) == 3 or die;

my $n_lo = 2**$level_lo;
my $n_aggregate = 2**$level_aggregate;
my $n_hi = 2**$level_hi;
my $x_min = n_to_x($n_lo);
my $x_aggregate = n_to_x($n_aggregate);
my $x_max = n_to_x($n_hi);

my $y_max = 0;
my $prev_y;
{
  # plot lines
  my $n_start = $n_lo;
  while ($n_start <= $n_aggregate && func($n_start)==0) {
    $prev_y = func($n_start);
    $n_start++;
  }
  my $x_start = n_to_x($n_start);
  $Printwrap = 2;
  printwrap("\\draw[line cap=butt] ");
  $Printwrap += 2;
  my $join = '';
  foreach my $n ($n_start .. $n_aggregate-1) {
    my $y = func($n);
    my $x = n_to_x($n);
    ### at: "n=$n x=$x prev_y=".($prev_y//'undef')." y=$y"

    # horizontal across to this X column
    if (defined $prev_y) {
      printwrap("$join($x,$prev_y)");
      $join = '--';
    }

    # vertical from prev Y to this Y in this column
    printwrap("$join($x,$y)");
    $join = '--';

    $prev_y = $y;
    $y_max = max($y,$y_max);
  }
  printwrap("--($x_aggregate,$prev_y)\n");
}

sub mask_to_aggregate {
  my ($n) = @_;
  my $s = 0;
  while ($n > $n_aggregate) {
    $n >>= 1;
    $s++;
  }
  return $n << $s;
}

{
  # plot aggregated

  my @n_list;
  my @yhi_list;
  my @ylo_list;
  foreach my $n ($n_aggregate .. $n_hi-1) {
    my $y = func($n);
    my $n_masked = mask_to_aggregate($n);
    ### put: "n=$n masked $n_masked  y=$y"
    if (@n_list && $n_masked==$n_list[-1]) {
      $yhi_list[-1] = max($yhi_list[-1], $y);
      $ylo_list[-1] = min($ylo_list[-1], $y);
    } else {
      push @n_list, $n_masked;
      push @yhi_list, $y;
      push @ylo_list, $y;
    }
  }

  # end the plot line down to the aggregate if necessary
  if ($prev_y < $ylo_list[0]) {
    printwrap("--($x_aggregate,$ylo_list[0]);\n");
  } elsif ($prev_y > $yhi_list[0]) {
    printwrap("--($x_aggregate,$yhi_list[0]);\n");
  } else {
    printwrap(";\n");
  }

  $y_max = max($y_max, @yhi_list);
  ### y last: "$ylo_list[-1] to $yhi_list[-1] step ".($yhi_list[-1] - $ylo_list[-1])
  print "  % total ",scalar(@n_list)," aggregated points\n";

  $prev_y = $ylo_list[0];
  $Printwrap = 2;
  printwrap("\\fill [shift={(line-width-lo)}] ($x_aggregate,$prev_y) ");
  $Printwrap += 2;
  foreach my $i (1 .. $#n_list) {
    my $n = $n_list[$i];
    my $y = $ylo_list[$i];
    my $x = n_to_x($n);
    printwrap(" --($x,$prev_y)");
    printwrap(" --($x,$y)");
    $prev_y = $y;
  }
  {
    my $x = n_to_x($n_hi);
    printwrap(" --($x,$prev_y)");
    printwrap(" [shift={(line-width-hi)}]");
    $prev_y = $yhi_list[-1];
    printwrap(" --($x,$prev_y)");
  }
  foreach my $i (reverse 1 .. $#n_list) {
    my $n = $n_list[$i];
    my $x = n_to_x($n);
    printwrap(" --($x,$prev_y)");
    my $y = $yhi_list[$i-1];
    printwrap(" --($x,$y)");
    $prev_y = $y;
  }
  printwrap(" --($x_aggregate,$prev_y)");
  printwrap(" --cycle;\n");
}

print "  \\coordinate (x-min) at ($x_min,0);\n";
print "  \\coordinate (x-max) at ($x_max,0);\n";
print "  \\coordinate (y-max) at (0,$y_max);\n";

# X ticks
$Printwrap = 2;
printwrap("\\draw");
$Printwrap += 2;
foreach my $k ($level_lo .. $level_hi) {
  my $n = 2**$k;
  my $x = n_to_x($n);
  my $semi = ($k==$level_hi ? ";\n" : "");
  printwrap(" ($x,0) -- +(x-tick)$semi");
}

# X numbers
print "  \\begin{scope}[font=\\scriptsize]\n";
foreach my $k ($level_lo .. $level_hi-1) {
  my $n = 2**$k;
  my $x = n_to_x($n);
  print "  \\node at ($x,0) [below right,my x number] {\$$n\$};\n";
}

print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
