#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use List::Util 'max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'rect_corners_sort','transform_to_horizontal',
  'xy_rotate_plus90','xy_rotate_minus90';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 20;

print "  % generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
my $scale = sqrt($x_hi**2 + $y_hi**2);
print "  \\begin{scope}\n";

sub scale {
  my ($x) = @_;
  $x = sprintf '%.6f', $x/$scale;
  $x =~ s/0+$//;
  return $x;
}
sub transform {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, $level);
  return (scale($x), scale($y));
}
sub untransform {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, -$level);
  return (int($x*$scale), int($y*$scale));
}

my ($tx_max,$ty_max) = (3/4 + .15, -1/4 + .052);
my ($tx_min,$ty_min) = (3/4 - .11, -1/4 - .075);
print "  \\coordinate (bottom-left) at ($tx_min,$ty_min);\n";
print "  \\coordinate (top-right) at ($tx_max,$ty_max);\n";

my ($x_min,$y_min) = untransform($tx_min,$ty_min);
my ($x_max,$y_max) = untransform($tx_max,$ty_max);
($x_min,$y_min, $x_max,$y_max) = rect_corners_sort($x_min,$y_min,
                                                   $x_max,$y_max);

sub xy_in_rectangle {
  my ($x,$y) = @_;
  return ($x_min <= $x && $x <= $x_max
          && $y_min <= $y && $y <= $y_max);
}
sub xy_on_rectangle_boundary {
  my ($x,$y) = @_;
  return ($x == $x_min || $x == $x_max
       || $y == $y_min || $y == $y_max);
}
sub xyxy_is_traversed {
  my ($x1,$y1, $x2,$y2) = @_;
  (xy_in_rectangle($x1,$y1) && xy_in_rectangle($x2,$y2)) or return 0;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2) // return 0;
  return 1;
}
sub xyxy_good {
  my ($x1,$y1, $x2,$y2) = @_;
  #### xyxy_good(): "$x1,$y1 to $x2,$y2"
  xyxy_is_traversed($x1,$y1, $x2,$y2) || return 0;

  if (xy_on_rectangle_boundary($x1,$y1)
      && xy_on_rectangle_boundary($x2,$y2)) {
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;
    my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
    foreach (1,2) {
      if (xyxy_is_traversed($x1,$y1, $x1+$rx, $y1+$ry)
          && xyxy_is_traversed($x2,$y2, $x2+$rx, $y2+$ry)) {
        return 1;
      }
      $rx = -$rx;
      $ry = -$ry;
    }
    return 0;
  }

  return 1;
}

# at $x,$y last step was in $dx,$dy
my %seen;
sub xydxdy_step {
  my ($x,$y, $dx,$dy) = @_;
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);
  foreach (1,2,3) {
    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    if (xyxy_good($x,$y, $x2,$y2)) {
      $seen{$x,$y} = 1;
      return ($x2,$y2, $dx,$dy);
    }
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);
  }
  return;
}

my $segments = MyLineSegments->new;
sub find_cycle {
  my ($x,$y, $dx,$dy) = @_;
  ### find_cycle(): "$x,$y dir $dx,$dy"
  my $orig_x = $x;
  my $orig_y = $y;
  my $length = 0;
  my ($x2,$y2);
  for (;;) {
    ### at: "$x,$y  $dx,$dy"
    unless (($x2,$y2, $dx,$dy) = xydxdy_step($x,$y, $dx,$dy)) {
      ### no further step ...
      if ($length != 0) {
        die "oops no step at $x,$y\n";
      }
      last;
    }
    $segments->add($x,$y, $x2,$y2);
    $length++;
    ($x,$y) = ($x2,$y2);
    if ($x == $orig_x && $y == $orig_y) {
      ### back at original: "$x,$y"
      last;
    }
  }
  # print "% length $length\n";
}

# print "  % sx range $sx_min to $sx_max\n";
foreach my $y ($y_min) {
  foreach my $x ($x_min .. $x_max) {   # bottom going right
    next if $seen{$x,$y};
    find_cycle($x,$y, 1,0);
  }
}
foreach my $y ($y_max) {
  foreach my $x (reverse $x_min .. $x_max) {   # top going left
    next if $seen{$x,$y};
    find_cycle($x,$y, -1,0);
  }
}

foreach my $x ($x_max) {
  foreach my $y ($y_min .. $y_max) {     # right going up
    next if $seen{$x,$y};
    find_cycle($x,$y, 0,1);
  }
}
foreach my $x ($x_min) {
  foreach my $y (reverse $y_min .. $y_max) {   # left going down
    next if $seen{$x,$y};
    find_cycle($x,$y, 0,-1);
  }
}

$Printwrap = 2;
printwrap("\\fill[my grey] ");
$Printwrap += 2;
$segments->printwrap (transform => \&transform);
printwrap(";");
printwrap(" % drawn with " . ($segments->lines_joined_count_points) . " points\n");

print "  \\end{scope}\n";
print "  % end generated\n";
exit 0;
