#!/usr/bin/perl -w

# Copyright 2019 Kevin Hyde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use FindBin;
use List::Util 'min','max';

use lib $FindBin::Bin;
use MyDragon 'bitlength','DirRatio';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

my $level_lo = 4;
my $level_aggregate = 7;
my $level_hi = 11;

print "  % generated by tools/$FindBin::Script\n";

my @DirRatioCumul = (0);
foreach my $n (1 .. 2**$level_hi) {
  $DirRatioCumul[$n] = DirRatio($n) + $DirRatioCumul[$n-1];   # inclusive
}

sub func {
  my ($n) = @_;
  return sprintf '%.6f', $DirRatioCumul[$n] / $n;
}

sub n_to_x {
  my ($n) = @_;
  if ($n==0) { return 0; }
  my $k = bitlength($n);
  my $high = 2**($k-1);
  my $r = $n - $high;
  return ($r ? "$k+$r/$high" : $k);
}
n_to_x(0) == 0 or die;
n_to_x(1) == 1 or die;
n_to_x(2) == 2 or die;
n_to_x(4) == 3 or die;


my $n_lo = 2**$level_lo;
my $x_min = n_to_x($n_lo);
my $n_aggregate = 2**$level_aggregate;
my $x_aggregate = n_to_x($n_aggregate);
my $y_max = 0;
my $prev_y = func($n_lo);
{
  # plot lines
  $Printwrap = 2;
  printwrap("\\draw[line cap=butt] ($x_min,$prev_y)");
  $Printwrap += 2;
  foreach my $n ($n_lo .. $n_aggregate-1) {
    my $y = func($n);
    my $x = n_to_x($n);
    printwrap("--($x,$prev_y)");
    printwrap("--($x,$y)");
    $prev_y = $y;
    $y_max = max($y,$y_max);
  }
  printwrap("--($x_aggregate,$prev_y)");
}

sub mask_to_aggregate {
  my ($n) = @_;
  my $s = 0;
  while ($n > $n_aggregate) {
    $n >>= 1;
    $s++;
  }
  return $n << $s;
}

my $n_hi = 2**$level_hi;
my $x_max = n_to_x($n_hi);
{
  # plot aggregated

  my @n_list;
  my @yhi_list;
  my @ylo_list;
  foreach my $n ($n_aggregate .. $n_hi-1) {
    my $y = func($n);
    my $n_masked = mask_to_aggregate($n);
    ### put: "n=$n masked $n_masked  y=$y"
    if (@n_list && $n_masked==$n_list[-1]) {
      $yhi_list[-1] = max($yhi_list[-1], $y);
      $ylo_list[-1] = min($ylo_list[-1], $y);
    } else {
      push @n_list, $n_masked;
      push @yhi_list, $y;
      push @ylo_list, $y;
    }
  }

  # end the plot line down to the aggregate if necessary
  if ($prev_y < $ylo_list[0]) {
    printwrap("--($x_aggregate,$ylo_list[0]);\n");
  } elsif ($prev_y > $yhi_list[0]) {
    printwrap("--($x_aggregate,$yhi_list[0]);\n");
  } else {
    printwrap(";\n");
  }

  $y_max = max($y_max, @yhi_list);
  ### y last: "$ylo_list[-1] to $yhi_list[-1] step ".($yhi_list[-1] - $ylo_list[-1])
  print "  % total ",scalar(@n_list)," aggregated points\n";

  $prev_y = $ylo_list[0];
  $Printwrap = 2;
  printwrap("\\fill [shift={(line-width-lo)}] ($x_aggregate,$prev_y) ");
  $Printwrap += 2;
  foreach my $i (1 .. $#n_list) {
    my $n = $n_list[$i];
    my $y = $ylo_list[$i];
    my $x = n_to_x($n);
    printwrap(" --($x,$prev_y)");
    printwrap(" --($x,$y)");
    $prev_y = $y;
  }
  {
    my $x = n_to_x($n_hi);
    printwrap(" --($x,$prev_y)");
    printwrap(" [shift={(line-width-hi)}]");
    $prev_y = $yhi_list[-1];
    printwrap(" --($x,$prev_y)");
  }
  foreach my $i (reverse 1 .. $#n_list) {
    my $n = $n_list[$i];
    my $x = n_to_x($n);
    printwrap(" --($x,$prev_y)");
    my $y = $yhi_list[$i-1];
    printwrap(" --($x,$y)");
    $prev_y = $y;
  }
  printwrap(" --($x_aggregate,$prev_y)");
  printwrap(" --cycle;\n");
}

print "  \\coordinate (x-min) at ($x_min, 1/2);\n";
print "  \\coordinate (x-max) at ($x_max, 1/2);\n";
print "  \\coordinate (y-max) at (0,$y_max);\n";

# X ticks
$Printwrap = 2;
printwrap("\\draw");
$Printwrap += 2;
foreach my $k ($level_lo+1 .. $level_hi) {
  my $n = 2**$k;
  my $x = n_to_x($n);
  my $semi = ($k==$level_hi ? ";\n" : "");
  printwrap(" ($x,1/2) -- +(x-tick)$semi");
}

# X numbers
print "  \\begin{scope}[font=\\scriptsize]\n";
foreach my $k ($level_lo .. $level_hi-1) {
  my $n = 2**$k;
  my $x = n_to_x($n);
  print "  \\node at ($x,1/2) [below right,my x number] {\$$n\$};\n";
}

print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
