#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Drawing downwards from the diameter, sub-trees directed left and down to
# make room for caption at right.

use 5.004;
use strict;
use List::Util 'min','max';
use Tie::IxHash;
use Graph::Maker::TwindragonAreaTree;

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'RQ','WidthS_infinity_k';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
my $max_depth = 33;
WidthS_infinity_k($max_depth+1) <= $level or die;
RQ($level-1) > $max_depth or die;

print "  % generated by tools/$FindBin::Script\n";

# $v1 and $v2 are vertices of $graph and $graph is a tree.
# $v1 is the parent of $v2
# Return the height (longest path to leaf) of the subtree at $v2.
sub Graph_subtree_height {
  my ($graph, $v1, $v2) = @_;
  ### $v1
  ### $v2
  $graph->has_edge($v1,$v2) or die "oops, $v1 and $v2 not adjacent";
  my $depth = 0;
  my %seen = ($v1 => 1, $v2 => 1);
  my @pending = ($v2);
  do {
    @pending = map {$graph->neighbours($_)} @pending;
    @pending = grep {! $seen{$_}++} @pending;
    $depth++;
  } while (@pending);
  return $depth;
}

my $graph = Graph::Maker->new('twindragon_area_tree', level=>$level,
                              undirected=>1);
my $root = 0;

my $segments = MyLineSegments->new;
my @vertices = ($root);
my %seen;
tie %seen, 'Tie::IxHash';
my $x = 0;
my @widths = (1);
my %vertex_to_location = ($root => [0,0]);
$seen{$root} = 1;
while (@vertices) {
  if (++$x > $max_depth) { last; }
  my $width = 0;
  my $y = 0;
  my @new_vertices;
  foreach my $v (@vertices) {
    my $v_location = $vertex_to_location{$v};
    my ($vx,$vy) = @$v_location;
    my @children = $graph->neighbours($v);
    ### $v
    ### @children
    @children = grep {! $seen{$_}} @children;
    @children = sort
      {Graph_subtree_height($graph,$v,$b) <=> Graph_subtree_height($graph,$v,$a)}
      @children;
    foreach my $child (@children) {
      $y = min($y, $vy);
      $vertex_to_location{$child} = [$x,$y];
      $segments->add($vx+$vy,$vy, $x+$y,$y);

      push @new_vertices, $child;
      $seen{$child} = 1;
      $y--;
      $width++;
    }
  }
  push @widths, $width;
  @vertices = @new_vertices;
}

{
  # lines
  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap (transform => sub { my ($x,$y) = @_;
                                           return ($x-$y, $y);
                                         });
  printwrap(";\n");
}

{
  # dashed lines at end
  $Printwrap = 2;
  printwrap("\\draw[my dashed] ");
  $Printwrap = 4;
  foreach my $v (@vertices) {
    my @children = $graph->neighbours($v);
    @children = grep {! $seen{$_}} @children;
    if (@children) {
      my $p = $vertex_to_location{$v};
      my ($x,$y) = @$p;
      printwrap(" ($x,$y)--+(\\MyDash,0)");
    }
  }
  printwrap(";\n");
}

{
  # dots
  $Printwrap = 2;
  printwrap("\\path foreach \\p in {");
  $Printwrap = 4;
  foreach my $v (keys %seen) {
    my $p = $vertex_to_location{$v};
    my ($x,$y) = @$p;
    printwrap("($x,$y),");
  }
  printwrap("}");
  printwrap(" { \\p node[my dot] };\n");
}

print "% GP-DEFINE  WidthS_infinity_generated_samples = [",
  join(',',@widths),"];\n";
print "  % end generated\n";
exit 0;
