#!/usr/bin/perl -w

# Copyright 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'Lpred_k_flat','Rpred','Rnext','Lpred_k','Lnext_k',
  'JN','BlobN', 'transform_to_horizontal';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 18;     # even
if (@ARGV) { $level = $ARGV[0]; }
print "  % generated by tools/$FindBin::Script $level\n";

my $path = Math::PlanePath::DragonCurve->new;
my $total_points = 0;

# Coordinates are put into MyLineSegments as integers with intended final
# scaling 1/2**$scale.
#
# TeX and PGF limit the size of integers, so 1/2**$extra_scale of this
# scaling is done as decimals in the output and "scale=..." capped.
#
my $scale = ($level+4)/2;
my $extra_scale = 0;
if ($scale > 9) {
  $extra_scale = $scale - 9;
  $scale = 9;
}
my $extra_scale_factor = 1/2**$extra_scale;
sub transform_extra_scale {
  my ($x,$y) = @_;
  return ($x*$extra_scale_factor, $y*$extra_scale_factor);
}

print "  \\begin{scope}[scale=1/2^$scale]\n";
{
  # Join
  my ($n_lo,$n_hi) = $path->level_to_n_range($level);
  my $n_start = JN($level);
  Lpred_k($level,$n_start) or die "oops";

  my $segments = MyLineSegments->new;
  my $join = '';
  my $count = 0;
  my $prev_cx;
  my $prev_cy;
  for (my $n = $n_start; defined $n; $n = Lnext_k($level,$n)) {
    ### assert: Lpred_k($level,$n)
    ### $n

    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;

    my $cx = ($x1+$x2 - $dy);
    my $cy = ($y1+$y2 + $dx);
    ($cx,$cy) = transform_to_horizontal($cx,$cy, $level+1);

    unless (defined $prev_cx && ($cx == $prev_cx && $cy == $prev_cy)) {
      $count++;  # not a repeat
    }
    if (defined $prev_cx) {
      $segments->add($prev_cx,$prev_cy, $cx,$cy);
    }
    $prev_cx = $cx;
    $prev_cy = $cy;
  }
  $Printwrap = 2;
  printwrap("\\draw[my grey] ");
  $Printwrap += 2;
  $segments->printwrap (transform => sub {
                          my ($x,$y) = @_;
                          return ($x*$extra_scale_factor, $y*$extra_scale_factor);
                        });
  printwrap(";\n");
  $total_points += $segments->lines_joined_count_points;

  print "% GP-Test  JA($level) == $count  /* drawn middles of squares */\n";
}

$level++;

{
  # Left of Blob
  my $n_start = BlobN($level)-1;
  my $n_end = BlobN($level+1) - 3;
  Lpred_k($level,$n_start) or die "oops";

  my $join = '';
  my $count = 0;
  my $prev_cx;
  my $prev_cy;
  my $segments = MyLineSegments->new;
  for (my $n = $n_start; $n <= $n_end; $n = Lnext_k($level,$n)) {
    ### assert: Lpred_k($level,$n)

    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;

    my $cx = ($x1+$x2 - $dy);
    my $cy = ($y1+$y2 + $dx);
    ($cx,$cy) = transform_to_horizontal($cx,$cy, $level);

    unless (defined $prev_cx && ($cx == $prev_cx && $cy == $prev_cy)) {
      $count++;  # not a repeat
    }
    if (defined $prev_cx) {
      $segments->add($prev_cx,$prev_cy, $cx,$cy);
    }
    $prev_cx = $cx;
    $prev_cy = $cy;
  }
  $Printwrap = 2;
  printwrap("\\draw[my grey] ");
  $Printwrap += 2;
  $segments->printwrap (transform => sub {
                          my ($x,$y) = @_;
                          return ($x*$extra_scale_factor, $y*$extra_scale_factor);
                        });
  printwrap(";\n");
  $total_points += $segments->lines_joined_count_points;

  print "% GP-Test  BlobLQ($level) == $count  /* drawn middles of squares */\n";
}

{
  # Right of Blob
  my $n_start = BlobN($level)-1;
  my $n_end = BlobN($level+1) - 3;
  Rpred($n_start) or die "oops, not right boundary";

  my $join = '';
  my $count = 0;
  my $prev_cx;
  my $prev_cy;
  my $segments = MyLineSegments->new;
  for (my $n = $n_start; $n <= $n_end; $n = Rnext($n)) {
    ### assert: Rpred($level,$n)

    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;

    my $cx = ($x1+$x2 + $dy);
    my $cy = ($y1+$y2 - $dx);
    ($cx,$cy) = transform_to_horizontal($cx,$cy, $level);

    unless (defined $prev_cx && ($cx == $prev_cx && $cy == $prev_cy)) {
      $count++;  # not a repeat
    }
    if (defined $prev_cx) {
      $segments->add($prev_cx,$prev_cy, $cx,$cy);
    }
    $prev_cx = $cx;
    $prev_cy = $cy;
  }
  $Printwrap = 2;
  printwrap("\\draw[my grey] ");
  $Printwrap += 2;
  $segments->printwrap (transform => sub {
                          my ($x,$y) = @_;
                          return ($x*$extra_scale_factor, $y*$extra_scale_factor);
                        });
  printwrap(";\n");
  $total_points += $segments->lines_joined_count_points;

  print "% GP-Test  BlobRQ($level) == $count  /* drawn middles of squares */\n";
}
print "  \\end{scope}\n";
print "  % total $total_points points\n";

print "  % end generated\n";
exit 0;


__END__

# sub next_str {
#   my ($strstate, $x,$y) = @_;
#   my $abs = "($x,$y)";
#   if (defined $strstate) {
#     my $dx = $x - $strstate->{'prev_x'};
#     my $dy = $y - $strstate->{'prev_y'};
#     my $rel = "+($dx,$dy)";
#     if (length($rel) < length($abs) && abs($dx)<10 && abs($dy)<10) {
#       return ($strstate, $rel);
#     }
#   }
#   return ({prev_x => $x, prev_y => $y}, $abs);
# }

