#!/usr/bin/perl -w

# Copyright 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Lingua::EN::Numbers;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'dir','BlobN','BlobSegments',
  'dir4_to_dx','dir4_to_dy','chamfer';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments '###';


my $level = 11;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $level\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;

# Return true if $x,$y is double-visited in the curve.
sub Dpred_xy {
  my ($x,$y) = @_;
  my @n_list = $path->xy_to_n_list($x,$y);
  return @n_list >= 2;
}

# $x,$y,$dir is a segment.
# Return true if it and following around by $side increments of $dir is a
# unit square of all 4 sides within $n_lo to $n_hi or $n2_lo to $n2_hi.
sub square_is_within {
  my ($x,$y, $dir, $n_lo,$n_hi,$side, $n2_lo,$n2_hi) = @_;
  foreach (1..4) {
    ### square_is_within: "$x,$y $dir"
    my $dx = dir4_to_dx($dir);
    my $dy = dir4_to_dy($dir);
    my $n_other = $path->xyxy_to_n_either($x,$y, $x+$dx, $y+$dy);
    unless (defined $n_other
            && (($n_other >= $n_lo && $n_other < $n_hi)
                || ($n_other >= $n2_lo && $n_other < $n2_hi))) {
      ### square not within ...
      return 0;
    }
    $x += $dx;
    $y += $dy;
    $dir += $side;
  }
  return 1;
}

# Return true if want chamfering at point $x,$y with next segment going in
# direction $dir mod 4.
sub want_chamfer_xy {
  my ($x,$y,$dir, $n_lo,$n_hi,$side, $n2_lo,$n2_hi) = @_;
  $n2_lo ||= 0;
  $n2_hi ||= 0;
  my @n_list = $path->xy_to_n_list($x,$y);

  if (@n_list < 2) {
    ### single-visited not chamfered ...
    return 0;
  }
  @n_list = grep {($n_lo <= $_ && $_ < $n_hi)
                    || ($n2_lo <= $_ && $_ < $n2_hi)} @n_list;
  if (@n_list < 2) {
    ### other visit not in range, want chamfer ...
    return 1;
  }
  return ! square_is_within($x,$y,$dir+2, $n_lo,$n_hi,-$side, $n2_lo,$n2_hi);
}

sub turn_to_within {
  my ($x,$y, $dir, $n_lo,$n_hi,$side, $not) = @_;
  $not ||= 0;
  # ### $x
  # ### $y
  # ### $dir
  # ### $n_lo
  # ### $n_hi
  foreach (1..4) {
    my $dx = dir4_to_dx($dir);
    my $dy = dir4_to_dy($dir);
    my $n_other = $path->xyxy_to_n_either($x,$y, $x+$dx, $y+$dy);
    if ((defined $n_other && $n_other >= $n_lo && $n_other < $n_hi)
        ^ $not) {
      return $dir;
    }
    $dir = ($dir+$side)&3;
  }
  die "oops, not found within";
}

sub boundary_list {
  my ($n_lo,$n_hi, $side, $n2_lo,$n2_hi) = @_;
  ### boundary_list(): "$n_lo to $n_hi side $side"
  ### $n2_lo
  ### $n2_hi

  my $n = $n_lo;
  my ($x,$y) = $path->n_to_xy($n);
  my $dir = turn_to_within($x,$y, 0, $n_lo,$n_hi,$side, 1);  # not within
  ### initial dir: $dir
  ### cf segment: dir($n)%4

  my @ret;
  my $prev_dir;
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
  ### target hi: "$x_hi,$y_hi"
  my %seen;
  for (;;) {
    $dir = turn_to_within($x,$y, $dir, $n_lo,$n_hi,$side);
    ### boundary_list at: "$x,$y dir=$dir"
    if ($seen{"$x,$y,$dir"}++) {
      die "oops, already seen";
    }

    my $dx = dir4_to_dx($dir);
    my $dy = dir4_to_dy($dir);
    ### want_chamfer: want_chamfer_xy($x,$y,$dir, $n_lo,$n_hi,$side, $n2_lo,$n2_hi)
    if (defined $prev_dir
        && !($x==$x_hi && $y==$y_hi)
        && want_chamfer_xy($x,$y,$dir, $n_lo,$n_hi,$side, $n2_lo,$n2_hi)
       ) {
      my $prev_dx = dir4_to_dx($prev_dir);
      my $prev_dy = dir4_to_dy($prev_dir);
      push @ret, "($x".chamfer(-$prev_dx) . ",$y".chamfer(-$prev_dy).")";
      push @ret, "($x".chamfer($dx) . ",$y".chamfer($dy).")";
    } else {
      push @ret, "($x,$y)";
    }
    $prev_dir = $dir;

    $x += $dx;
    $y += $dy;
    $dir = ($dir-$side)&3;  # following around the side

    if ($x==$x_hi && $y==$y_hi) {
      last;
    }
  }
  return @ret;
}

sub points_concat {
  my ($aref1, $aref2) = @_;
  if (@$aref1 && @$aref2
      && $aref1->[-1] eq $aref2->[0]) {
    pop @$aref1;
  }
  if (@$aref1 && @$aref2
      && $aref2->[-1] eq $aref1->[0]) {
    pop @$aref2;
  }
  return (@$aref1, @$aref2);
}

my $n_lo = BlobN($level);
{
  my ($x,$y) = $path->n_to_xy($n_lo-1);
  print "  \\coordinate (start) at ($x,$y);\n";
}
my $n_hi = BlobN($level+1) - 3;
{
  my ($x,$y) = $path->n_to_xy($n_hi+1);
  print "  \\coordinate (end) at ($x,$y);\n";
}

sub boundary_list_right_and_left {
  my ($n_lo, $n_hi) = @_;
  my @right = boundary_list($n_lo,$n_hi,1);
  ### right points: scalar(@right)
  ### @right
  my ($x,$y) = $path->n_to_xy($n_lo);
  my ($x_end,$y_end) = $path->n_to_xy($n_hi);
  my @left;
  unless ($x==$x_end && $y==$y_end) {
    @left = reverse boundary_list($n_lo,$n_hi, -1);
    ### left points: scalar(@left)
    ### left reversed: @left
  }
  return points_concat(\@right,\@left);
}

my @lines;
my $blob_mid_n_lo;
my $blob_mid_n_hi;

# blobs
if ($level >= 8) {
  $Printwrap = 2;
  printwrap("\\path[my blob style]\n");
  $Printwrap += 2;
  {
    my $n = $n_lo;
    for (my $k = $level-4; $k >= 4; $k -= 2) {
      push @lines, [$n - ($k==$level-4 ? 1 : 3), $n];
      my $len = BlobSegments($k);
      my $end = $n + $len;
      ### $k
      ### $n
      ### $end
      printwrap("% n $n to $end length $len\n");
      my @points = boundary_list_right_and_left($n,$end);
      my $join = '';
      foreach my $p (@points) {
        printwrap("$join$p");
        $join = '--';
      }
      printwrap("${join}cycle\n");

      $n = $end + 3;
    }
    $blob_mid_n_lo = $n - 3;
  }
  {
    my $end = $n_hi;
    for (my $k = $level-4; $k >= 4; $k -= 2) {
      push @lines, [$end, $end + ($k==$level-4 ? 1 : 3)];
      my $len = BlobSegments($k);
      my $n = $end - $len;
      ### $k
      ### $n
      ### $end
      printwrap("% n $n to $end length $len\n");
      my @points = boundary_list_right_and_left($n,$end);
      my $join = '';
      foreach my $p (@points) {
        printwrap("$join$p");
        $join = '--';
      }
      my $semi = ($k-2 >= 4 ? '' : ';');
      printwrap("${join}cycle$semi\n");

      $end = $n - 3;
    }
    $blob_mid_n_hi = $end + 3;
  }
}

# twindragons
{
  my $twin_first_n2;
  my @k_list_first;
  {
    # from unfold midpoint
    my $n1 = my $n2 = 2**($level-1);
    for (my $k = $level-4; $k >= 0; $k -= 4) {
      ### k first: $k

      push @k_list_first, $k;
      my $n1_len = 2**$k;
      my $n2_len = 3*$n1_len;
      my $n1_lo = $n1 - $n1_len;
      my $n2_hi = $n2 + $n2_len;
      {
        my ($x1,$y1) = $path->n_to_xy($n1);
        my ($x2,$y2) = $path->n_to_xy($n2);
        unless ($x1==$x2 && $y1==$y2) {
          die "oops, different n1,n2 firsts";
        }
      }
      {
        my ($x1,$y1) = $path->n_to_xy($n1_lo);
        my ($x2,$y2) = $path->n_to_xy($n2_hi);
        unless ($x1==$x2 && $y1==$y2) {
          die "oops, different n1,n2 seconds";
        }
      }

      my @points1 = boundary_list($n1_lo,$n1, 1, $n2,$n2_hi);   # right
      my @points2 = boundary_list($n2,$n2_hi, 1, $n1_lo,$n1);   # right
      $Printwrap = 2;
      printwrap("% twin firsts  n1 $n1_lo to $n1 length $n1_len,  n2 $n2 to $n2_hi length $n2_len\n");
      printwrap("\\path[my twin $k style] ");
      $Printwrap += 2;
      my $join = '';
      foreach my $p (points_concat(\@points1,\@points2)) {
        printwrap("$join$p");
        $join = '--';
      }
      printwrap("${join}cycle;\n");

      $n1 = $n1_lo;
      $n2 = $n2_hi;
    }
    ### $blob_mid_n_lo
    ### $blob_mid_n_hi
    ### $n1
    ### $n2
    push @lines, [$blob_mid_n_lo, $n1];
    $twin_first_n2 = $n2;
  }

  my @k_list_second;
  {
    # from 3/4
    my $n1 = my $n2 = 3*2**($level-2);
    for (my $k = $level-6; $k >= 0; $k -= 4) {
      ### k second: $k

      push @k_list_second, $k;
      my $n2_len = 2**$k;
      my $n1_len = 3*$n2_len;
      my $n1_lo = $n1 - $n1_len;
      my $n2_hi = $n2 + $n2_len;
      {
        my ($x1,$y1) = $path->n_to_xy($n1);
        my ($x2,$y2) = $path->n_to_xy($n2);
        unless ($x1==$x2 && $y1==$y2) {
          die "oops, different n1,n2 firsts";
        }
      }
      {
        my ($x1,$y1) = $path->n_to_xy($n1_lo);
        my ($x2,$y2) = $path->n_to_xy($n2_hi);
        unless ($x1==$x2 && $y1==$y2) {
          die "oops, different n1,n2 seconds";
        }
      }

      my @points1 = boundary_list($n1_lo,$n1, -1, $n2,$n2_hi);  # left
      my @points2 = boundary_list($n2,$n2_hi, -1, $n1_lo,$n1);  # left
      $Printwrap = 2;
      printwrap("% twin seconds  n1 $n1_lo to $n1 length $n1_len,  n2 $n2 to $n2_hi length $n2_len\n");
      printwrap("\\path[my twin $k style] ");
      $Printwrap += 2;
      my $join = '';
      foreach my $p (points_concat(\@points1,\@points2)) {
        printwrap("$join$p");
        $join = '--';
      }
      printwrap("${join}cycle;\n");

      $n1 = $n1_lo;
      $n2 = $n2_hi;
    }
    ### $blob_mid_n_lo
    ### $blob_mid_n_hi
    ### $n1
    ### $n2
    push @lines, [$n2, $blob_mid_n_hi];
    push @lines, [$twin_first_n2, $n1];
  }

  foreach my $i (0 .. $#k_list_first) {
    my $k = $k_list_first[$i];
    my $num = ucfirst(Lingua::EN::Numbers::num2en($i+1));
    print "  \\newcommand{\\MyTwindragonFirst$num}{$k}\n";
  }
  foreach my $i (0 .. $#k_list_second) {
    my $k = $k_list_second[$i];
    my $num = ucfirst(Lingua::EN::Numbers::num2en($i+1));
    print "  \\newcommand{\\MyTwindragonSecond$num}{$k}\n";
  }
}


sub lines_is_other_endpoint {
  my ($n, $line) = @_;
  my ($x,$y) = $path->n_to_xy($n);
  foreach my $other_line (@lines) {
    next if $line == $other_line;
    foreach my $other_n (@$other_line) {
      my ($other_x,$other_y) = $path->n_to_xy($other_n);
      if ($other_x==$x && $other_y==$y) {
        ### other end: "n=$n $x,$y line=".join(',',@$line)." other $other_n $other_x,$other_y other_line=".join(',',@$other_line)
        return 1;
      }
    }
  }
  return 0;
}

# lines
{
  $Printwrap = 2;
  printwrap("\\draw\n");
  $Printwrap += 2;
  my %line_endpoints = (map {$_ => 1} @lines);
  foreach my $i (0 .. $#lines) {
    my $line = $lines[$i];
    ### $line
    my ($n_lo, $n_hi) = @$line;
    printwrap("% $n_lo to $n_hi\n");
    my $join = ' ';
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      my $end = ($i==$#lines && $n==$n_hi ? ";\n"
                 : $n==$n_hi ? "\n"
                 : '');
      if (lines_is_other_endpoint($n, $line)
          || ($n > $n_lo && $n < $n_hi
              && Dpred_xy($x,$y, $n_lo,$n_hi))) {
        my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
        my ($tdx,$tdy) = $path->n_to_dxdy($n);
        my $sx = $x . chamfer(-$sdx);
        my $sy = $y . chamfer(-$sdy);
        my $tx = $x . chamfer($tdx);
        my $ty = $y . chamfer($tdy);
        printwrap("$join($sx,$sy)");
        printwrap("--($tx,$ty)$end");
      } else {
        printwrap("$join($x,$y)$end");
      }
      $join = '--';
    }
  }
}

print "  % end generated\n";
exit 0;
