#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Generate curve segments surrounding the k=112 bits of x containing ffixed.
# The output is only small, but the bignum crunching of 112 bits is a touch
# slow.

use 5.010;
use strict;
use warnings;
use FindBin;
use List::Util 'min','max';
use Math::BigInt try => 'GMP';
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyLineSegments;
use MyDragon;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


print "  % generated by tools/$FindBin::Script\n";

my $path = Math::PlanePath::DragonCurve->new;

# j=4, 112 bits
my $x_bits = '1101000011111101000000000000110100000000000000000000000011010000111111111111111111111111111111111111111111111101';

# $x_bits = '11010000111111010000000000001101';    # j=2, 32 bits
# $x_bits = '0000';

my $x_segment  = Math::BigInt->new("0b$x_bits");
my $k = length($x_bits);
my ($n_lo, $n_hi) = $path->level_to_n_range(2*$k);

print "  \\node (x) at (0,0) [my dot];\n";

# within $n_hi
sub is_traversed {
  my ($x1,$y1, $x2,$y2) = @_;
  my $n = $path->xyxy_to_n_either ($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_hi;
}


my $segments = MyLineSegments->new;
my $x_shifted_lo = -4;
my $x_shifted_hi =  4;
my $y_lo         = -2;
my $y_hi         = 2;
my $x_lo = $x_segment + $x_shifted_lo;
my $x_hi = $x_segment + $x_shifted_hi;

sub is_xy_in_range {
  my ($x,$y) = @_;
  return ($x_lo <= $x && $x <= $x_hi
          && $y_lo <= $y && $y <= $y_hi);
}

my $scale = 4;
my $x_scaled_lo = $x_shifted_lo * $scale - 1;
my $x_scaled_hi = $x_shifted_hi * $scale + 1;
my $y_scaled_lo = $y_lo         * $scale - 1;
my $y_scaled_hi = $y_hi         * $scale + 1;
sub xy_scaled_truncate {
  my ($x,$y) = @_;
  return (max($x_scaled_lo, min($x_scaled_hi, $x)),
          max($y_scaled_lo, min($y_scaled_hi, $y)));
}

sub add_segment {
  my ($x,$y, $x2,$y2) = @_;
  ### add_segment() ...
  $x  -= $x_segment;     # so not bignums
  $x2 -= $x_segment;
  $x  = $x->numify;
  $x2 = $x2->numify;
  ### shifted: "$x,$y to $x2,$y2"
  $x *= $scale;
  $y *= $scale;
  $x2 *= $scale;
  $y2 *= $scale;
  ### scaled: "$x,$y to $x2,$y2"
  ($x,$y)   = xy_scaled_truncate($x,$y);
  ($x2,$y2) = xy_scaled_truncate($x2,$y2);
  ### truncated: "$x,$y to $x2,$y2"
  $segments->add ($x,$y, $x2,$y2);
}

for (my $x = $x_lo - 1; $x <= $x_hi; $x++) {
  for (my $y = Math::BigInt->new($y_lo - 1); $y <= $y_hi; $y++) {
    {
      # horizontal to the right at $x,$y
      my $x2 = $x+1;
      my $y2 = $y;
      if (is_xy_in_range($x,$y) || is_xy_in_range($x2,$y2)) {
        ### consider: "$x,$y to $x2,$y2"
        if (is_traversed($x,$y, $x2,$y2)) {
          add_segment($x,$y, $x2,$y2);
        }
      }
    }
    {
      # vertical upwards at $x,$y
      my $x2 = $x;
      my $y2 = $y+1;
      if (is_xy_in_range($x,$y) || is_xy_in_range($x2,$y2)) {
        ### consider: "$x,$y to $x2,$y2"
        if (is_traversed($x,$y, $x2,$y2)) {
          add_segment($x,$y, $x2,$y2);
        }
      }
    }
  }
}

$Printwrap = 2;
printwrap("\\draw ");
$Printwrap +=2;
$segments->printwrap (transform => sub {
                        my ($x,$y) = @_;
                        $x /= $scale;
                        $y /= $scale;
                        return ($x,$y);
                      });
printwrap(";\n");

print "  % end generated\n";
exit 0;
