#!/usr/bin/perl -w

# Copyright 2015 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use File::Spec;
use File::stat;
use File::Slurp;
use POSIX ();
use Graph;
use List::Util 'min','max';

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyDragon 'Graph_layout_diameterline';
use MyPrintwrap ':all';
use Graph::Maker::TwindragonAreaTree;
use Image::Base::GD;

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
my $filename = "/tmp/twindragon-area-tree-$level-line.png";

my $scale = 3;
my $background_colour = 'black';
my $line_colour = '#'.('B0'x3);
my $dot_colour = 'white';

# my $background_colour = 'white';
# my $line_colour = '#D0D0D0';
# my $dot_colour = 'black';

my $graph = Graph::Maker->new('twindragon_area_tree',
                              level       => $level,
                              undirected  => 1,
                              graph_maker => sub {
                                return Graph::Hashed->new(@_);
                              },
                             );
my $layout;
{
  local $SIG{'__WARN__'} = sub {
    return if $_[0] =~ /^Deep recursion/;
    warn @_;
  };
  $layout = Graph_layout_diameterline($graph);
}

my $x_min = min(map {$_->[0]} values %$layout);
my $x_max = max(map {$_->[0]} values %$layout);
my $x_width = $x_max - $x_min;

my $y_min = min(map {$_->[1]} values %$layout);
my $y_max = max(map {$_->[1]} values %$layout);
my $y_height = $y_max - $y_min;

my $border = 10;
my $top_border = 20;
my $bottom_border = 10;

### $x_width
### $y_height
# $scale = max(1, int(min($scale, ($width/$x_width), ($height/$y_height))));

my $width  = $scale*$x_width  + 1 + 2*$border;
my $height = $scale*$y_height + 1 + $bottom_border + $top_border;

my $image = Image::Base::GD->new (-height => $height,
                                  -width  => $width);
$image->rectangle(0,0, $width-1,$height-1, $background_colour);

my $transform = sub {
  my ($x,$y) = @_;
  $x -= $x_min;
  $y -= $y_min;
  $x *= $scale;
  $y *= $scale;
  $x += $border;
  $y += $bottom_border;
  $y = $height-1 - $y;  # up the screen
  return ($x,$y);
};

foreach my $edge ($graph->edges) {
  my ($v1,$v2) = @$edge;
  my ($x1,$y1) = @{$layout->{$v1}} or die;
  my ($x2,$y2) = @{$layout->{$v2}} or die;
  $image->line($transform->($x1,$y1),
               $transform->($x2,$y2),
               $line_colour);
}

foreach my $v ($graph->vertices) {
  my ($x,$y) = @{$layout->{$v}} or die;
  ($x,$y) = $transform->($x,$y);
  $image->xy($x,$y, $dot_colour);
}

$image->save($filename);

# add text to the png image in $filename
sub pngtextadd {
  my ($filename, $keyword, $value) = @_;
  system('pngtextadd', "--keyword=$keyword", "--text=$value", $filename) == 0
    or die "system(pngtextadd)";
}
{
  my $program_filename = File::Spec->catfile($FindBin::Bin,
                                             $FindBin::Script);
  my $copyright = File::Slurp::read_file($program_filename);
  $copyright =~ /^# (Copyright.*)/m or die;
  $copyright = $1;

  # PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
  use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
  my $stat = File::stat::stat($program_filename)
    or die "Oops, cannot stat $program_filename: $!";
  my $creation_time = POSIX::strftime(STRFTIME_FORMAT_RFC822,
                                      localtime($stat->mtime));

  pngtextadd($filename, 'Author',    'Kevin Ryde');
  pngtextadd($filename, 'Generator', "tools/$FindBin::Script");
  pngtextadd($filename, 'Title',     "Twindragon area tree level $level, diameter layout");
  pngtextadd($filename, 'Creation Time', $creation_time);
  pngtextadd($filename, 'Copyright', $copyright);
  pngtextadd($filename, 'Homepage',  'http://user42.tuxfamily.org/dragon/index.html');

  print "uncompressed ",-s $filename,"\n";
  require IPC::Run;
  IPC::Run::run(['optipng','-quiet','-o5',$filename]);
  print "compressed   ",-s $filename,"\n";
}

exit 0;
